/*----------------------------------------------------------------------
 *  AbstractDomNode
 *----------------------------------------------------------------------
 */

function AbstractDomNode(domNode, keepDomNode) {
    this.children = [];
    this.images = [];
    
    this._extractCharacteristics(domNode, keepDomNode);
    
    for (var i = 0; i < domNode.childNodes.length; i++) {
        var childNode = domNode.childNodes[i];
        if ((childNode.nodeType == 1) || 
            (childNode.nodeType == 3 && trim(childNode.nodeValue).length > 0)) {
            this.children.push(new AbstractDomNode(childNode, keepDomNode));
        }
    }
    
    if (this.children.length > 0) {
        this._aggregateCharacteristics();
    }
    
    if (keepDomNode) {
        this.domNode = domNode;
    }
}

AbstractDomNode.prototype._createAttributeNode = function(name, value, domNode, keepDomNode) {
    return {
        children :          [],
        images :            [],
        nodeType :          2,
        nodeName :          "@" + name,
        text :              value,
        cssClass :          "",
        displayBlock :      false,
        domNode :           keepDomNode ? domNode : null,
        computeEditCost :   
            AbstractDomNode.prototype.computeEditCost,
        computeEditCostIncludeChildren :   
            AbstractDomNode.prototype.computeEditCostIncludeChildren
    };
}

AbstractDomNode.prototype._extractCharacteristics = function(domNode, keepDomNode) {
    this.nodeType = domNode.nodeType;
    
    this.text = "";
    this.cssClass = "";
    this.displayBlock = false;
    
    if (domNode.nodeType == 1) { // element node
        this.nodeName = domNode.nodeName;
        this.cssClass = domNode.className;
        
        var tagName = domNode.tagName.toUpperCase();
        if (":P:DIV:TABLE:TBODY:TR:TD:UL:OL:LI:".indexOf(":" + tagName + ":") >= 0) {
            this.displayBlock = true;
        }
        
        if (tagName == "IMG") {
            this.children.push(this._createAttributeNode("src", domNode.src, domNode, keepDomNode));
            this.images.push(domNode.src);
        } else if (tagName == "A") {
            //this.children.push(this._createAttributeNode("href", domNode.href, domNode, keepDomNode));
        }
        
        if ((domNode.title) && domNode.title.length > 0) {
            this.children.push(this._createAttributeNode("title", domNode.title, domNode, keepDomNode));
        }
    } else if (domNode.nodeType == 3) { // text node
        this.text = trim(domNode.nodeValue);
        this.nodeName = "text";
    }
}

AbstractDomNode.prototype._aggregateCharacteristics = function() {
    if (this.children.length == 1 && this.children[0].displayBlock) {
        this.displayBlock = true;
    }
    
    var textFragments = [];
    for (var i = 0; i < this.children.length; i++) {
        var childADN = this.children[i];
        if (childADN.nodeType != 2 && childADN.text.length > 0) {
            textFragments.push(childADN.text);
        }
        if (childADN.images.length > 0) {
            this.images = this.images.concat(childADN.images);
        }
    }
    
    this.text = textFragments.join(" ");
}

AbstractDomNode.prototype.computeEditCost = function(adn) {
    var cost = 0;
    if (this.nodeType != adn.nodeType && (this.nodeType == 2 || adn.nodeType == 2)) {
        cost = Number.POSITIVE_INFINITY;
    } else {
        if (this.nodeName != adn.nodeName) {
            cost += 0.5;
        }
        if (this.cssClass != adn.cssClass) {
            cost += 0.25;
        }
        if (this.displayBlock != adn.displayBlock) {
            cost += 0.25;
        }
    }
    return cost;
}

AbstractDomNode.prototype.computeEditCostIncludeChildren = function(adn) {
    var thisCost = this.computeEditCost(adn);
    var cost = 0;
    
    var us = 0;
    var them = 0;
    
    while (us < this.children.length && them < adn.children.length) {
        var ourChild = this.children[us];
        var theirChild = adn.children[them];
        
        var theirBestOffset = 0;
        var theirMinCost = 1000;
        for (var i = them; i < adn.children.length; i++) {
            var c = ourChild.computeEditCost(adn.children[i]) * (theirBestOffset + 1);
            if (c < theirMinCost) {
                theirBestOffset = i - them;
                theirMinCost = c;
            }
        }
        
        var ourBestOffset = 0;
        var ourMinCost = 1000;
        for (var i = us; i < this.children.length; i++) {
            var c = theirChild.computeEditCost(this.children[i]) * (ourBestOffset + 1);
            if (c < ourMinCost) {
                ourBestOffset = i - us;
                ourMinCost = c;
            }
        }
        
        if (theirMinCost < ourMinCost) {
            // We should insert their nodes into ours
            cost += theirBestOffset;
            
            us += 1;
            them += theirBestOffset + 1;
        } else {
            // We should insert our nodes into theirs
            cost += ourBestOffset;
            
            us += ourBestOffset + 1;
            them += 1;
        }
    }
    
    if (us < this.children.length) {
        cost += (this.children.length - us);
    } else if (them < adn.children.length) {
        cost += (adn.children.length - them);
    }
    
    var factor = (this.children.length + adn.children.length + 1);
    return (thisCost * factor + cost) / (factor + this.children.length + adn.children.length);
}
