
/*
 * The Piggy Bank object skeleton
 */
var PiggyBank = new Object();

/*
 * Error information that gets populated if problems happen during startup
 */
PiggyBank._error = "";
PiggyBank._error_location = ""

/*
 * The browser XUL has loaded. Find the elements we need and set up our
 * listeners and wrapper objects.
 */
PiggyBank.chromeLoad = function() {

    // NOTE: we don't trace 'chromeload' because the logger is not
    // initialized anyway and will never get logged, also to avoid
    // confusing the indentation level, we don't log the exit either.

    PiggyBank._initialize();

    // we setup a progress listener to hook to the 'location change' events
    // that we use to activate the data coin icon on the page status bar

    PiggyBank._listener = PB_BrowserUtilities.createProgressListener();
    PiggyBank._listener.onLocationChange = function (aProgress, aRequest, aStatus) {
        PB_Debug.trace("piggybank.js","> PiggyBank.onLocationChange");
        PiggyBank._instrumentBrowsers();
        PB_Debug.trace("piggybank.js","< PiggyBank.onLocationChange");
    }

    PiggyBank._tabBrowser.addProgressListener(PiggyBank._listener, Components.interfaces.nsIWebProgress.NOTIFY_LOCATION);

    // the meat of the piggybank browsing instrumentation happens when the page
    // has finished loading and the DOM tree is in place, so that we can
    // react on the RDF that is already contained/linked or to the URL of
    // the page matching the regular expressions of some installed scraper
    // NOTE: due to the slowness of the RDF data extraction and parsing,
    // we don't do anything to the data until you click the 'data coin' icon.

    PiggyBank._onPageLoad = function(aEvent) {
        PB_Debug.trace("piggybank.js","> PiggyBank.onPageLoad");
        PiggyBank._loadingInterceptor(PiggyBank._runCollector);
        PB_Debug.trace("piggybank.js","< PiggyBank.onPageLoad");
    };

    // FIXME(SM): how can we get called after all dom-modification firefox extensions
    // (say greasemonkey) have finished manipulating the dom? Is extension order in the
    // extension pane influencing this?

    document.getElementById("appcontent").addEventListener("load", PiggyBank._onPageLoad, true);

	// save the previous URI content listener so that we can call 
	// it later to handle N3 files coming out of Piggy Bank
	try {
		var docShell = PiggyBank._tabBrowser.docShell;
		PiggyBank._contentListener =
			docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
				.getInterface(Components.interfaces.nsIURIContentListener);
	} catch (ex) {
	}
		
    // now we need to attach our content listener to be invoked when the browser loads
    // a URI that returns some RDF content that we can work with directly
    Components.classes["@mozilla.org/uriloader;1"]
        .getService(Components.interfaces.nsIURILoader)
            .registerContentListener(PiggyBank);

}

/*
 * The browser XUL has unloaded. Remove listeners to avoid leaking memory.
 */
PiggyBank.chromeUnload = function() {
    PB_Debug.trace("piggybank.js","> chromeUnload");

    var server = PB_Extension.getPiggyBankServer();
    if (server) {
        for (var i = 0; i < PiggyBank._browsers.length; i++) {
            var browser = PiggyBank._browsers[i];
            var sourceName = PiggyBank._browserToSourceName[browser];
            if (sourceName) {
                // tell the java subsystem to discard the temporary profiles
                server.removeSource(sourceName);
            }
            // dispose the collector (which contains listerners!) to avoid leaking memory
            PiggyBank._getCollector(browser).dispose();
        }
    }

    // remove the listeners to avoid leaking memory
    window.getBrowser().removeProgressListener(PiggyBank._listener);
    document.getElementById("appcontent").removeEventListener("load", PiggyBank._onPageLoad, true);

    // unregister as a URI content listener
    Components.classes["@mozilla.org/uriloader;1"]
        .getService(Components.interfaces.nsIURILoader)
            .unRegisterContentListener(PiggyBank);

    // cut the references to the objects we created
    PiggyBank._browsers = null;
    PiggyBank._browserToCollector = null;
    PiggyBank._browserToSourceName = null;
	PiggyBank._contentListener = null;
    PiggyBank._tabBrowser = null;
    PiggyBank._listener = null;

    PB_Debug.trace("piggybank.js","< chromeUnload");
    PB_Debug.trace("piggybank.js","----- [Disposed PiggyBank Overlay] -----");
}

// ---------------------------------------------------------------------------------

PiggyBank._loadingInterceptor = function(continuation) {
    PB_Debug.trace("piggybank.js","> PiggyBank._loadingInterceptor");

    if (!PB_Extension.isRunning()) { // if piggybank is not running
        if (!PB_Extension.isLoading()) { // and piggybank not already loading
            setTimeout(function() {
                PiggyBank._load(continuation); // then start it
            }); // but it in another thread
                // so that we don't block the browser
                // the load function will call the continuation
                // as soon as it has finished loading
        }
        // The piggy bank subsystem initialization is now happening in the background
        // this means that until it's completed the data icon won't show up
        // but we return so that the user can still use the browser.

        // This improves the user perceived browser startup performance
        // and prevents the user from thinking that piggybank has crashed their
        // system when, for whatever reason (first run, slow to load the JVM,
        // migrating and reindexing data from previous versions, etc..),
        // it takes a while to start.
    } else {
        continuation();
    }

    PB_Debug.trace("piggybank.js","< PiggyBank._loadingInterceptor");
}

PiggyBank._instrumentBrowsers = function() {
    PB_Debug.trace("piggybank.js","> PiggyBank.instrumentBrowsers");

    var browsers = PiggyBank._tabBrowser.browsers;

    PB_Debug.trace("piggybank.js","> Remove Collectors");

    // Remove collector of any browser that no longer exists
    var x = 0;
    while (x < PiggyBank._browsers.length) {
        var browser = PiggyBank._browsers[x];
        var exists = false;

        for (var j = 0; j < browsers.length; j++) {
            if (browser == browsers[j]) {
                exists = true;
                break;
            }
        }

        if (!exists) {
            PiggyBank._browsers.splice(x,1);

            var sourceName = PiggyBank._browserToSourceName[browser];
            if (sourceName) {
                PB_Extension.getPiggyBank().getPiggyBankServer().removeSource(sourceName);
                PiggyBank._browserToSourceName[browser] = null;
            }

            PiggyBankExplorePane.onBrowserClosed(browser);
            
            PiggyBank._removeCollector(browser).dispose();
        } else {
            x++;
        }
    }

    PB_Debug.trace("piggybank.js","< Remove Collectors");

    PB_Debug.trace("piggybank.js","> Add Collectors");

    // Add a collector to any new browser
    for (var i = 0; i < browsers.length; i++) {
        var browser = browsers[i];
        var exists = false;

        for (var j = 0; j < PiggyBank._browsers.length; j++) {
            if (browser == PiggyBank._browsers[j]) {
                exists = true;
                break;
            }
        }

        if (!exists) {
            PiggyBank._browsers.splice(i,0,browser);
            PiggyBank._setCollector(browser, new PB_Collector(PiggyBank._tabBrowser, browser));
        }
    }

    PB_Debug.trace("piggybank.js","< Add Collectors");
    
    var selectedBrowser = PiggyBank._tabBrowser.selectedBrowser;

    PiggyBank._getCollector(selectedBrowser).updateWindow();
    PiggyBankExplorePane.onBrowserSelected(selectedBrowser);

    PB_Debug.trace("piggybank.js","< PiggyBank.instrumentBrowsers");
}

PiggyBank._runCollector = function() {
    PB_Debug.trace("piggybank.js","> PiggyBank._runCollector");
    PiggyBank._getCollector(PiggyBank._tabBrowser.selectedBrowser).loadDataForCurrentDocument();
    PB_Debug.trace("piggybank.js","< PiggyBank._runCollector");
}

// ---------------------------------------------------------------------------------

PiggyBank._initialize = function() {

    try {

        // obtain the required xpcom components
        var log = Components.classes["@simile.mit.edu/logger;1"].getService(Components.interfaces.nsILogger)
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);

        // Make sure that the pig toolbar icon (if present!) indicates we are currently disabled
        Pig.setStatus("status","disabled");

        // Prepare the browser and collector instrumentation caches --------------------
        PiggyBank._browsers = [];
        PiggyBank._browserToCollector = [];
        PiggyBank._browserToSourceName = [];
        PiggyBank._tabBrowser = PB_BrowserUtilities.getTabBrowser();

        // Process the debug and tracing parameters ------------------------------------

        PiggyBank._debug = false;

        // Debug Mode
        var prefType = prefs.getPrefType(PB_Preferences.debug);
        if (prefType == Components.interfaces.nsIPrefBranch.PREF_BOOL) {
            PiggyBank._debug = prefs.getBoolPref(PB_Preferences.debug);
            log.enable(PiggyBank._debug);
        } else {
            prefs.setBoolPref(PB_Preferences.debug, PiggyBank._debug);
        }

        // Make sure that the pig toolbar icon (if present!) is in debug mode
        Pig.setStatus("debug",PiggyBank._debug);

        // JavaScript tracing
        var prefType = prefs.getPrefType(PB_Preferences.debug_tracing_javascript);
        if (prefType == Components.interfaces.nsIPrefBranch.PREF_BOOL) {
            log.enableTracing(prefs.getBoolPref(PB_Preferences.debug_tracing_javascript) && PiggyBank._debug);
        } else {
            prefs.setBoolPref(PB_Preferences.debug_tracing_javascript, false);
        }

        PB_Debug.trace("piggybank.js","----- [New PiggyBank Overlay] -----");
        PB_Debug.trace("piggybank.js","> initialize");

        // Redirect the javascript console to the system console
        var prefType = prefs.getPrefType(PB_Preferences.debug_redirectJSConsole);
        if (prefType == Components.interfaces.nsIPrefBranch.PREF_BOOL) {
            log.redirectJSConsole(prefs.getBoolPref(PB_Preferences.debug_redirectJSConsole) && PiggyBank._debug);
        } else {
            prefs.setBoolPref(PB_Preferences.debug_redirectJSConsole, false);
        }

        PB_Debug.trace("piggybank.js","< initialize");

    } catch (e) {
        PiggyBank._error = e.name + " - " + e.message + " - " + PB_Debug.introspect(e);
        PiggyBank._error_location = "initialization";
        Pig.setStatus("status","disabled");
        throw e; // make sure this gets logged too
    }
}

PiggyBank._load = function (continuation) {

	var showDoc = false;
	var showButton = false;

    try {
        PB_Debug.trace("piggybank.js","> load");

        var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);

        // Make sure java is available and works for our needs

        if (!PB_BrowserUtilities.isJavaAvailable()) {
            if (prefs.getPrefType(PB_Preferences.java_not_found_dialog_shown) != Components.interfaces.nsIPrefBranch.PREF_BOOL) {
                prefs.setBoolPref(PB_Preferences.java_not_found_dialog_shown, false);
            }
            var javaNotFoundDialogShown = prefs.getBoolPref(PB_Preferences.java_not_found_dialog_shown);
            if (!javaNotFoundDialogShown) {
                prefs.setBoolPref(PB_Preferences.java_not_found_dialog_shown, true);
                var macosx = navigator.userAgent.indexOf("Mac OS X") != -1;
                if (macosx) {
                    PiggyBank.showJEPNotFoundPage();
                } else {                    
                    PiggyBank.showJavaNotFoundPage();
                }
            }
            PB_Debug.trace("piggybank.js","< load");
            return;
        } else {
            if (prefs.getPrefType(PB_Preferences.java_not_found_dialog_shown) == Components.interfaces.nsIPrefBranch.PREF_BOOL) {
                prefs.clearUserPref(PB_Preferences.java_not_found_dialog_shown);
            }
        }

        // Make sure that the pig toolbar icon (if present!) indicates we are loading
        Pig.setStatus("status","loading");

        // Signal that we are loading, so that we avoid initializing more than one time

        PB_Extension.setLoading(true);

        // NOTE(SM): the java and java classloading traces will be written to the
        // java console, not the JavaScript console, nor the STDOUT.
        // I was able to modify the java plugin on MacOS to redirect the java console
        // into STDOUT (which then is redirected to the System Console), but
        // unfortunately it's not something I can't control from here
        // (or, at least, I haven't found a way to do it)

        // Understand if piggybank was ever run and what version was run before --------

        var previous_version = null;
        
        if (prefs.getPrefType(PB_Preferences.version) == Components.interfaces.nsIPrefBranch.PREF_STRING) {
            previous_version = prefs.getCharPref(PB_Preferences.version);
        } else {
            // if no version is found, this could mean that piggybank was never
            // installed on the system or that piggybank 2.0.x was found before.
            // So, we check that and transform the deprecated property used
            // by previous piggybanks to indicate their existance into the new
            // version-aware format
            if (prefs.getPrefType(PB_Preferences.initialize_deprecated) == Components.interfaces.nsIPrefBranch.PREF_BOOL) {
                if (!prefs.getBoolPref(PB_Preferences.initialize_deprecated)) {
                    previous_version = "2.0.x";
                    prefs.setCharPref(PB_Preferences.version, previous_version);
                }
                prefs.clearUserPref(PB_Preferences.initialize_deprecated)
            } else {
                // if we get here, piggybank was never installed before
                previous_version = "";
            }
        }

        // WARNING (SM): we do *NOT* set the version at the beginning but at the very
        // end of the initialization process, when we are sure everything went well,
        // otherwise if the browser crashes during the first startup we find
        // ourselves in an inconsistent state we'll never be able to get out
        // without manual tweaking of the preferences (which we don't want!)
        
        if (PB_Extension.showWizard(previous_version)) {
        	// start with UI feedbacks
        	var results = {
        	   showDoc : false,
        	   showButton : false
            }
            window.openDialog(
            	"chrome://piggy-bank/content/welcome-wizard.xul",
            	"welcome-dialog",
            	"chrome,dialog,modal,width=600,height=380",
            	previous_version,
            	results
            );
            showDoc = results.showDoc;
            showButton = results.showButton;
        } else {
        	// start without UI feedbacks
        	var progress = {
        		taskCompleted : function() {},
        		done : function() {},
        		fail : function(e) {}
        	};
	        PiggyBank.start(progress, previous_version);
	    }

        // record that the installation was successful ---------------------------------

        if (PB_Extension.isRunning()) {
            prefs.setCharPref(PB_Preferences.version, PB_Extension.getVersion());
            Pig.setStatus("status","ready");
        } else {
            Pig.setStatus("status","disabled");
            PiggyBank._error_location = "loading";
            PiggyBank._error = PB_Extension.getError();
        }

        PB_Extension.setLoading(false); // record that we finished loading

        PB_Debug.trace("piggybank.js","< load");
    } catch (e) {
        PiggyBank._error = e.name + " - " + e.message;
        PiggyBank._error_location = "loading";
        Pig.setStatus("status","disabled");
        throw e; // make sure this gets logged too
    }

    // if present, call the continuation
    if (continuation) continuation();

    // show documentation
    if (showDoc) PB_BrowserUtilities.loadURI("http://simile.mit.edu/piggy-bank/whats-next.html");

    // show the toolbar button
    if (showButton) setTimeout(window.BrowserCustomizeToolbar,100); // this shows the toolbar customize dialog

    //if (showButton) PiggyBank.placeButton(); // this places the button in the toolbar 
                                               // but, for some reason, it makes firefox *very* unstable

}

PiggyBank.start = function(progress, previous_version) {

    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
    
    // Retrieve the port the internal web server will listen to --------------------

    var port = PB_Extension.default_port;

    if (prefs.getPrefType(PB_Preferences.port) == Components.interfaces.nsIPrefBranch.PREF_INT) {
        port = prefs.getIntPref(PB_Preferences.port);
    } else {
        // if it wasn't set, set it (we need it in the options dialog)
        prefs.setIntPref(PB_Preferences.port, port);
    }

    // Retrieve the ID for generating the tag URIs ---------------------------------

    var idForGeneratingTagURIs = "";

    if (prefs.getPrefType(PB_Preferences.idForGeneratingTagURIs) == Components.interfaces.nsIPrefBranch.PREF_STRING) {
        idForGeneratingTagURIs = prefs.getCharPref(PB_Preferences.idForGeneratingTagURIs);
    }

    // Retrieve the ID to enable Google Maps ---------------------------------------

    var googleKey = "";

    if (prefs.getPrefType(PB_Preferences.googleKey) == Components.interfaces.nsIPrefBranch.PREF_STRING) {
        googleKey = prefs.getCharPref(PB_Preferences.googleKey);
    } else {
        prefs.setCharPref(PB_Preferences.googleKey, googleKey);
    }

    // Retrieve the class locations (for development purposes) ---------------------

    if (prefs.getPrefType(PB_Preferences.debug_classes_longwell) == Components.interfaces.nsIPrefBranch.PREF_STRING) {
        var longwellClasses = prefs.getCharPref(PB_Preferences.debug_classes_longwell);
    }

    if (prefs.getPrefType(PB_Preferences.debug_classes_piggybank) == Components.interfaces.nsIPrefBranch.PREF_STRING) {
        var piggybankClasses = prefs.getCharPref(PB_Preferences.debug_classes_piggybank);
    }

    // get the global debug mode
    if (prefs.getPrefType(PB_Preferences.debug) == Components.interfaces.nsIPrefBranch.PREF_BOOL) {
        var debug = prefs.getBoolPref(PB_Preferences.debug);
    }
    
    // Java Classloader tracing
    if (prefs.getPrefType(PB_Preferences.debug_tracing_classloader) == Components.interfaces.nsIPrefBranch.PREF_BOOL) {
        var traceClassloader = prefs.getBoolPref(PB_Preferences.debug_tracing_classloader) && debug;
    } else {
        prefs.setBoolPref(PB_Preferences.debug_tracing_classloader, false);
    }

    // Java tracing
    if (prefs.getPrefType(PB_Preferences.debug_tracing_java) == Components.interfaces.nsIPrefBranch.PREF_BOOL) {
        var traceJava = prefs.getBoolPref(PB_Preferences.debug_tracing_java) && debug;
    } else {
        prefs.setBoolPref(PB_Preferences.debug_tracing_java, false);
    }

    // initialize the java side of things ------------------------------------------

    PB_Extension.initialize(
        java,
        previous_version,
        port,
        PiggyBank._getExtensionPath(PB_Extension.getPB_ExtensionName()),
        PiggyBank._getProfilePath(),
        idForGeneratingTagURIs,
        googleKey,
        traceClassloader,
        traceJava,
        piggybankClasses,
        longwellClasses,
        progress
    );
}
    
// this function places the piggybank toolbar button in the toolbar
// NOTE(SM): this appears to make the browser extremely unstable!! so we don't use it for now
PiggyBank.placeButton = function() {

    // find search field
    var searchfield = document.getElementById("search-container");

    // Address bar found and button not already in a toolbar
    if ((searchfield != null) && (document.getElementById("piggyBank-toolbarButton") == null)) {

        // get address bar's toolbar
        var toolbarNode = searchfield.parentNode;

        // add your button in front of search field
        if (toolbarNode.insertItem("piggyBank-toolbarButton",searchfield,null,false)) {

            // get toolbar's currentset and remove all instances of your button if already stored
            var currentset = toolbarNode.currentSet.replace(/piggyBank-toolbarButton,?/gi,"");

            // remove end of line comma is there is one
            if (currentset.charAt(currentset.length-1) == ",") currentset = currentset.substring(0, currentset.length - 1);

            // add your button before address bar in currentset and make it persistent
            currentset = currentset.replace(/search-container/i,"piggyBank-toolbarButton,search-container");
            toolbarNode.setAttribute("currentset",currentset);
            toolbarNode.currentSet = currentset;
            document.persist(toolbarNode.id,"currentset");
        }
    }
}
    
// ---------------------------------------------------------------------------------

PiggyBank.checkBeforeExecuting = function(continuation) {
    if (!PB_Extension.attemptedToLoad()) {
        PiggyBank._load();
    }
    if (PB_Extension.isRunning()) {
        continuation();
    } else {
        PiggyBank.showErrorPage(PiggyBank._error, PiggyBank._error_location);
    }
}

PiggyBank.showJavaNotFoundPage = function() {
    window.openDialog("chrome://piggy-bank/content/javanotfound-dialog.xul", "javaNotFoundDialog", "chrome,dialog");
}

PiggyBank.showJEPNotFoundPage = function() {
    window.openDialog("chrome://piggy-bank/content/jepnotfound-dialog.xul", "jepNotFoundDialog", "chrome,dialog");
}

PiggyBank.showErrorPage = function(error,location) {
    window.openDialog("chrome://piggy-bank/content/error-dialog.xul", "errorDialog", "chrome,modal,dialog", error, location);
}

// ---------------------------------------------------------------------------------

PiggyBank.reload = function() {
    PB_Debug.trace("piggybank.js","> PiggyBank.reload");

    // dispose the backend
    PB_Extension.dispose();

    // dispose the frontend
    PiggyBank.chromeUnload();

    // reinitialize the frontend
    PiggyBank.chromeLoad();

    // reload the frontend
    PiggyBank._loadingInterceptor(PiggyBank._runCollector);

    PB_Debug.trace("piggybank.js","< PiggyBank.reload");
}

PiggyBank.browseTidbits = function() {
    PB_Debug.trace("piggybank.js","> PiggyBank.browseTidbits");
    PiggyBank._getCollector(PiggyBank._tabBrowser.selectedBrowser).browseTidbits();
    PB_Debug.trace("piggybank.js","< PiggyBank.browseTidbits");
}

PiggyBank.browseMyPiggyBank = function() {
    PB_Debug.trace("piggybank.js","> PiggyBank.browseMyPiggyBank");
    Pig.setStatus("status","working");
    PB_Extension.browseProfile("default", PiggyBank._tabBrowser, PiggyBank._tabBrowser.selectedBrowser);
    PB_Debug.trace("piggybank.js","< PiggyBank.browseMyPiggyBank");
}

PiggyBank.track = function() {
    PB_Debug.trace("piggybank.js","> PiggyBank.track");

    var browser = PiggyBank._tabBrowser.selectedBrowser;
    var piggyBank = PB_Extension.getPiggyBank();
    var server = piggyBank.getPiggyBankServer();
    var oldName = browser.getAttribute("piggyBank-sourceName");

    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
        .getService(Components.interfaces.nsIPromptService);

    var result = { value : oldName };

    while (promptService.prompt(
            window,
            "Piggy Bank",
            "Name this window or tab",
            result,
            null,
            { value : result.value })) {

        if (server.hasSource(result.value)) {
            promptService.alert(window, "Piggy Bank", "This name has already been used.");
        } else if (!(result.value) || result.value.length == 0) {
            promptService.alert(window, "Piggy Bank", "Name cannot be blank.");
        } else {
            if (oldName) {
                server.removeSource(oldName);
            }

            browser.setAttribute("piggyBank-sourceName", result.value);
            PiggyBank._browserToSourceName[browser] = result.value;

            server.setSource(result.value, browser.contentDocument.location.href);

            break;
        }
    }

    PB_Debug.trace("piggybank.js","< PiggyBank.track");
}

// -----------------------------------------------------------------------------------------------------

PiggyBank._removeCollector = function(browser) {
    PB_Debug.trace("piggybank.js","> PiggyBank._removeCollector");
    var collector = PiggyBank._browserToCollector[PiggyBank._getBrowserKey(browser)];
    PiggyBank._browserToCollector[PiggyBank._getBrowserKey(browser)] = null;
    PB_Debug.trace("piggybank.js","< PiggyBank._removeCollector");
    return collector;
}

PiggyBank._setCollector = function(browser, collector) {
    PB_Debug.trace("piggybank.js","PiggyBank._setCollector");
    PiggyBank._browserToCollector[PiggyBank._setBrowserKey(browser)] = collector;
}

PiggyBank._getCollector = function(browser) {
    PB_Debug.trace("piggybank.js","PiggyBank._getCollector");
    var key = PiggyBank._getBrowserKey(browser);
    return PiggyBank._browserToCollector[key];
}

PiggyBank._getBrowserKey = function(browser) {
    PB_Debug.trace("piggybank.js","PiggyBank._getBrowserKey");
    return browser.getAttribute("piggyBank-key");
}

PiggyBank._setBrowserKey = function(browser) {
    PB_Debug.trace("piggybank.js","PiggyBank._setBrowserKey");
    var key = (new Date()).getTime();
    browser.setAttribute("piggyBank-key", key);
    return key;
}

PiggyBank._getExtensionPath = function(extensionName) {
    PB_Debug.trace("piggybank.js","> getPB_ExtensionPath");
    var chromeRegistry =
        Components.classes["@mozilla.org/chrome/chrome-registry;1"]
            .getService(Components.interfaces.nsIChromeRegistry);
    var uri =
        Components.classes["@mozilla.org/network/standard-url;1"]
            .createInstance(Components.interfaces.nsIURI);

    uri.spec = "chrome://" + extensionName + "/content/";

    var path = chromeRegistry.convertChromeURL(uri);
    if (typeof(path) == "object") path = path.spec
    path = path.substring(0, path.indexOf("/chrome/") + 1);

    PB_Debug.trace("piggybank.js","< getPB_ExtensionPath");
    return path;
}

PiggyBank._getProfilePath = function() {
    PB_Debug.trace("piggybank.js","> getProfilePath");
    var fileLocator =
        Components.classes["@mozilla.org/file/directory_service;1"]
            .getService(Components.interfaces.nsIProperties);

    var path = escape(fileLocator.get("ProfD", Components.interfaces.nsIFile).path.replace(/\\/g, "/")) + "/";
    if (path.indexOf("/") == 0) {
    	path = 'file://' + path;
    } else {
    	path = 'file:///' + path;
    }

    PB_Debug.trace("piggybank.js","< getProfilePath");
    return path;
}

// ----------------------- The part implementing the content listener ---------------------

PiggyBank.QueryInterface = function(iid) {
    if (iid.equals(Components.interfaces.nsIURIContentListener) ||
        iid.equals(Components.interfaces.nsISupportsWeakReference) ||
        iid.equals(Components.interfaces.nsISupports))
    {
        return PiggyBank;
    } else {
        throw Components.results.NS_NOINTERFACE;
    }
}

PiggyBank.onStartURIOpen = function(uri) {
    PB_Debug.trace("piggybank.js","ContentListener:onStartURIOpen(" + uri + ")");
    return false;
}

PiggyBank.doContent = function(contentType, isContentPreferred, request, contentHandler) {
    PB_Debug.trace("piggybank.js","ContentListener:doContent(" + request.URI.spec + " [" + contentType + "])");
    if (request.URI.spec.indexOf("http://127.0.0.1") < 0) {
        PB_Debug.trace("piggybank.js","ContentListener: Content from remotely");
        PiggyBank._loadingInterceptor(function() {
            PiggyBank._getCollector(PiggyBank._tabBrowser.selectedBrowser).loadDataFromURL(request);
            PiggyBank.browseTidbits();
        });
        return true;
    } else {
        PB_Debug.trace("piggybank.js","ContentListener: Content local");
        
        if (PiggyBank._contentListener)
			// do not pass contentHandler along or it won't work
			return PiggyBank._contentListener.doContent(contentType, isContentPreferred, request);
		else
			return false;
    }
}

PiggyBank.isPreferred = function(contentType, desiredContentType) {
    PB_Debug.trace("piggybank.js","ContentListener:isPreferred(" + contentType + "," + desiredContentType + ")");
    return PiggyBank.canHandleContent(contentType);
}

const RDF_XML_MIME_TYPE = "application/rdf+xml";
const RDF_XML_MIME_TYPE_DEPRECATED = "text/rdf";
const RDF_N3_MIME_TYPE = "text/rdf+n3";
const RDF_N3_MIME_TYPE_DEPRECATED = "application/n3";
const RDF_TURTLE_MIME_TYPE = "application/turtle";
const RDF_TURTLE_MIME_TYPE_UNOFFICIAL = "application/x-turtle";

PiggyBank.canHandleContent = function(contentType, isContentPreferred, desiredContentType) {
    var canHandle =
        (contentType == RDF_XML_MIME_TYPE) ||
        (contentType == RDF_XML_MIME_TYPE_DEPRECATED) ||
        (contentType == RDF_N3_MIME_TYPE) ||
        (contentType == RDF_N3_MIME_TYPE_DEPRECATED) ||
        (contentType == RDF_TURTLE_MIME_TYPE) ||
        (contentType == RDF_TURTLE_MIME_TYPE_UNOFFICIAL);
    PB_Debug.trace("piggybank.js","ContentListener:canHandleContent(" + contentType + ") -> " + canHandle);
    return canHandle;
}

PiggyBank.loadCookie = null;

PiggyBank.parentContentListener = null;

// -----------------------------------------------------------------------------------------------------

PiggyBank.init = function() {
    window.addEventListener("load", PiggyBank.chromeLoad, false);
    window.addEventListener("unload", PiggyBank.chromeUnload, false);
}
