function TreeNode(dataNode) {
    this.dataNode = dataNode;
    this.dataNode["_treeNode_"] = this;
    
    this.opened = false;
    this.previouslyOpened = false;
    this.constructedUI = false;
}

TreeNode.prototype = {
    get hasChildren() {
        return this.dataNode.hasChildren;
    }
};

TreeNode.prototype.dispose = function() {
    if (this.children) {
        for (var i = 0; i < this.children.length; i++) {
            this.children[i].dispose();
        }
    }
    this.dataNode["_treeNode_"] = null;
}

TreeNode.prototype.close = function() {
    if (this.children) {
        for (var i = 0; i < this.children.length; i++) {
            this.children[i].close();
        }
    }
    this.opened = false;
}

TreeNode.prototype.countOccupiedRows = function() {
    var count = 1;
    if (this.opened) {
        if (this.children) {
            for (var i = 0; i < this.children.length; i++) {
                count += this.children[i].countOccupiedRows();
            }
        }
    }
    return count;
}

function TreeColumn(o) {
    this.id = o.id;
    this.hidden = (o.hidden);
}

function RichTree(headers, dataNodes) {
    this._headers = headers;
    
    this._treeNodes = [];
    for (var i = 0; i < dataNodes.length; i++) {
        this._treeNodes.push(new TreeNode(dataNodes[i]));
    }
    
    this._constructTable();
}

RichTree.prototype = {
    _XHTMLNS: "http://www.w3.org/1999/xhtml",
    get element() {
        return this._table;
    }
};

RichTree.prototype._constructTable = function() {
    var table = document.createElementNS(this._XHTMLNS, "html:table");
    table.cellSpacing = 0;
    table.cellPadding = 0;
    this._table = table;
    
    var headerRow = this._constructRow(0);
    for (var i = 0; i < this._headers.length; i++) {
        var header = this._headers[i];
        var cell = this._constructHeaderCell(headerRow, i);
        
        var div = document.createElementNS(this._XHTMLNS, "html:div");
        div.className = "padded";
        div.appendChild(document.createTextNode(header.label));
        
        cell.appendChild(div);
    }
    
    for (var i = 0; i < this._treeNodes.length; i++) {
        this._insertRow(this._treeNodes[i], i+1);
    }
    
    for (var i = 0; i < this._treeNodes.length; i++) {
        this._openTreeNode(this._treeNodes[i]);
    }
};

RichTree.prototype._toggleTreeNode = function(treeNode) {
    if (treeNode.opened) {
        this._closeTreeNode(treeNode);
    } else {
        this._openTreeNode(treeNode);
    }
}

RichTree.prototype._openTreeNode = function(treeNode) {
    if (treeNode.opened) {
        return;
    }
    
    var parentNode = this._table.parentNode;
    var nextSibling;
    if (parentNode) {
        nextSibling = this._table.nextSibling;
        parentNode.removeChild(this._table);
    }
    
    treeNode.opened = true;
    treeNode.previouslyOpened = true;
    
    var dataNode = treeNode.dataNode;
    var rowIndex = this._findRowIndex(treeNode);
    
    var row = this._table.rows[rowIndex];
    this._destructCells(row);
    this._constructCells(treeNode, row);
    
    if (!treeNode.children) {
        treeNode.children = [];
        
        var childDataNodes = dataNode.children;
        for (var i = 0; i < childDataNodes.length; i++) {
            treeNode.children.push(new TreeNode(childDataNodes[i]));
        }
    }
    
    for (var i = 0; i < treeNode.children.length; i++) {
        this._insertRow(treeNode.children[i], rowIndex + i + 1);
    }
    
    if (treeNode.children.length == 1 && treeNode.children[0].hasChildren) {
        this._openTreeNode(treeNode.children[0]);
    } else {
        for (var i = 0; i < treeNode.children.length; i++) {
            if (treeNode.children[i].previouslyOpened) {
                this._openTreeNode(treeNode.children[i]);
            }
        }
    }
    
    if (parentNode) {
        if (nextSibling) {
            parentNode.insertBefore(this._table, nextSibling);
        } else {
            parentNode.appendChild(this._table);
        }
    }
};

RichTree.prototype._closeTreeNode = function(treeNode) {
    if (!treeNode.opened) {
        return;
    }
    
    var parentNode = this._table.parentNode;
    var nextSibling;
    if (parentNode) {
        nextSibling = this._table.nextSibling;
        parentNode.removeChild(this._table);
    }
    
    var totalRows = treeNode.countOccupiedRows();
    
    treeNode.close();
    treeNode.previouslyOpened = false;
    
    var rowIndex = this._findRowIndex(treeNode);
    for (var i = 0; i < totalRows; i++) {
        this._table.deleteRow(rowIndex);
    }
    this._insertRow(treeNode, rowIndex);
    
    if (parentNode) {
        if (nextSibling) {
            parentNode.insertBefore(this._table, nextSibling);
        } else {
            parentNode.appendChild(this._table);
        }
    }
}

RichTree.prototype._findRowIndex = function(treeNode) {
    var path = this._getAncestorTreeNodes(treeNode);
    
    var row = 0; // don't need to include header row as algorithm already adds 1 extra
    var treeNodes = this._treeNodes;
    for (var i = 0; i < path.length; i++) {
        if (!(treeNodes)) {
            return -1;
        }
        
        var pathTreeNode = path[i];
        
        var found = false;
        for (var j = 0; j < treeNodes.length; j++) {
            if (treeNodes[j] == pathTreeNode) {
                row += 1;
                treeNodes = treeNodes[j].children;
                found = true;
                break;
            } else {
                row += treeNodes[j].countOccupiedRows();
            }
        }
        
        if (!found) {
            return -1;
        }
    }
    
    return row;
};

RichTree.prototype._insertRow = function(treeNode, index) {
    var row = this._constructRow(index);
    this._constructCells(treeNode, row);
};

RichTree.prototype._constructCells = function(treeNode, row) {
    var tree = this;
    for (var j = 0; j < this._headers.length; j++) {
        var header = this._headers[j];
        var cell = this._constructCell(row, j);
        if (j == 0) {
            /*
             *  Create tree lines and open/close controls
             */
            var hbox = document.createElement("hbox");
            hbox.className = "tree-controls";
            
            var path = this._getAncestorTreeNodes(treeNode);
            var siblings = this._treeNodes;
            
            for (var k = 0; k < path.length; k++) {
                var n = path[k];
                
                var hasSibling = false;
                for (var s = 0; s < siblings.length; s++) {
                    if (siblings[s] == n) {
                        hasSibling = (s < siblings.length - 1);
                        siblings = n.children;
                        break;
                    }
                }
                
                var div = document.createElementNS(this._XHTMLNS, "html:div");
                div.className = hasSibling ? "tree-has-sibling" : "tree-no-sibling";
                
                if (k == path.length - 1) {
                    var rootOrNode = "root"; //(k == 0) ? "root" : "node";
                    var src = "chrome://solvent/skin/tree-" + rootOrNode + "-" + 
                        (n.hasChildren ? (n.opened ? "opened" : "closed") : "leaf") +
                        ".png";
                        
                    var image = document.createElementNS(this._XHTMLNS, "html:img");
                    image.src = src;
                    image.onclick = function() { tree._toggleTreeNode(treeNode); };
                    if (n.hasChildren) {
                        image.className = "tree-node tree-open-box";
                    } else {
                        image.className = "tree-node";
                    }
                    
                    div.appendChild(image);
                }
                
                hbox.appendChild(div);
            }
            
            var div = document.createElementNS(this._XHTMLNS, "html:div");
            hbox.appendChild(div);
            
            treeNode.dataNode.constructCell(div, header.id, treeNode.opened);
            
            cell.appendChild(hbox);
        } else {
            treeNode.dataNode.constructCell(cell, header.id, treeNode.opened);
        }
    }
};

RichTree.prototype._destructCells = function(row) {
    for (var j = this._headers.length - 1; j >= 0; j--) {
        row.deleteCell(j);
    }
};

RichTree.prototype._getTreeNode = function(dataNode) {
    return dataNode["_treeNode_"];
};

RichTree.prototype._getAncestorTreeNodes = function(treeNode) {
    var path = [];
    var treeNode2 = treeNode;
    while (treeNode2 != null) {
        path.unshift(treeNode2);
        var parentDataNode = treeNode2.dataNode.parent;
        if (parentDataNode) {
            treeNode2 = this._getTreeNode(parentDataNode);
        } else {
            treeNode2 = null;
        }
    }
    return path;
};

RichTree.prototype._constructRow = function(index) {
    return this._table.insertRow(index);
};

RichTree.prototype._constructHeaderCell = function(row, index) {
    var cell = row.insertCell(index);
    cell.className = "tree-cell tree-header";
    
    return cell;
};

RichTree.prototype._constructCell = function(row, index) {
    var cell = row.insertCell(index);
    if (index == 0) {
        cell.className = "tree-cell tree-controls-cell";
    } else {
        cell.className = "tree-cell";
    }
    return cell;
};