/*----------------------------------------------------------------------
 *  TemplateNode
 *----------------------------------------------------------------------
 */
 
function TemplateNode(sampleADN, children) {
    this.sampleADN = sampleADN;
    this.children = children;
    this.optional = false;
    this.hits = 0;
    this.misses = 0;
}

TemplateNode.createTemplateNode = function(adn) {
    var children = map(
        function(adnNode) {
            return new TemplateNode(adnNode, null);
        },
        adn.children
    );
    return new TemplateNode(adn, children);
}

TemplateNode.prototype.fillChildren = function(adn) {
    this.children = map(
        function(adnNode) {
            return new TemplateNode(adnNode, null);
        },
        adn.children
    );
}

TemplateNode.prototype.learnChildren = function(adn) {
    if (this.children == null) {
        throw new Error("Internal error");
    }
    
    var costFunction = function(templateNode, adn) {
        return templateNode.sampleADN.computeEditCostIncludeChildren(adn);
    };
    var compareFunction = function(templateNode, adn) {
        return (templateNode.sampleADN.nodeName != adn.nodeName) ? 1 : 0;
    };
    var hitFunction = function(templateNode) {
        templateNode.hits++;
    };
    var missFunction = function(templateNode) {
        templateNode.optional = true;
        templateNode.misses++;
    };
    var constructorFunction = function(adn) {
        return new TemplateNode(adn, null);
    };
    
    Template._learn(
        this.children, 
        adn.children, 
        costFunction, 
        compareFunction, 
        hitFunction, 
        missFunction, 
        constructorFunction
    );
}

TemplateNode.prototype.learnDeep = function(adns) {
    var start = 0;
    
    if (this.children == null) {
        while (start < adns.length) {
            if (adns[start]) {
                this.fillChildren(adns[start]);
                start++;
                break;
            }
            start++;
        }
    }
    if (this.children == null) {
        this.children = [];
        return;
    }
    
    for (var i = start; i < adns.length; i++) {
        if (adns[i]) {
            this.learnChildren(adns[i]);
        }
    }
    
    for (var j = 0; j < adns.length; j++) {
        if (adns[j]) {
            this.bindChildren(adns[j]);
        }
    }
    
    var adnIndices = map(function(adn) { return 0; }, adns);
    for (var i = 0; i < this.children.length; i++) {
        var childTemplateNode = this.children[i];
        var childADNS = [];
        
        for (var j = 0; j < adns.length; j++) {
            var childADN = null;
            
            if (adns[j]) {
                if (adnIndices[j] < adns[j].children.length) {
                    var candidateChildADN = adns[j].children[adnIndices[j]];
                    if (candidateChildADN.binding && candidateChildADN.binding.templateNodeIndex == i) {
                        childADN = candidateChildADN;
                        adnIndices[j]++;
                    }
                }
            }
            
            childADNS.push(childADN);
        }
        
        if (childTemplateNode.sampleADN.nodeType == 1) {
            childTemplateNode.learnDeep(childADNS);
        } else {
            childTemplateNode.learnText(childADNS);
        }
    }
}

TemplateNode.prototype.assignExtractableFields = function(param, level) {
    if (this.sampleADN.nodeType == 1) {
        if (this.children.length > 0) {
            var hitRate = 0;
            for (var i = 0; i < this.children.length; i++) {
                var child = this.children[i];
                hitRate += child.hits / (child.hits + child.misses);
            }
            hitRate /= this.children.length;
            
            if (hitRate < 0.5) {
                this.fieldName = "field" + param.fieldIndex;
                param.fieldIndex = param.fieldIndex + 1;
            } else {
                for (var i = 0; i < this.children.length; i++) {
                    this.children[i].assignExtractableFields(param, level + 1);
                }
            }
        }
    } else {
        var variableCount = 0;
        for (var i = 0; i < this.children.length; i++) {
            variableCount += this.children[i].variable ? 1 : 0;
        }
        if (variableCount > this.children.length * 2 / 3) {
            this.fieldName = "field" + param.fieldIndex;
            param.fieldIndex = param.fieldIndex + 1;
        } else {
            for (var i = 0; i < this.children.length; i++) {
                var textTemplateNode = this.children[i];
                if (textTemplateNode.variable) {
                    textTemplateNode.fieldName = "field" + param.fieldIndex;
                    param.fieldIndex = param.fieldIndex + 1;
                }
            }
        }
    }
}

TemplateNode.prototype.bindChildren = function(adn, extractFunction, extractTextFunction) {
    /*
     *  The trick here is that the template should always
     *  has "more" children than the adn (considering
     *  wildcards).
     */
     
    var us = 0;
    
    for (var them = 0; them < adn.children.length; them++) {
        var theirADN = adn.children[them];
        var match = false;
        
        var bestOffset = 0;
        var minCost = 1000;
        
        for (var i = us; i < this.children.length; i++) {
            var ourTemplateNode = this.children[i];
            var ourADN = ourTemplateNode.sampleADN;
            
            var c = theirADN.computeEditCostIncludeChildren(ourADN);
            if (ourTemplateNode.optional) {
                c = c / (1 + i - us);
            }
            
            if (c < minCost) {
                bestOffset = i - us;
                minCost = c;
            }
            
            if (!ourTemplateNode.optional) {
                break;
            }
        }
        
        us += bestOffset;
        
        var binding = {
            templateNode : this.children[us],
            templateNodeIndex : us
        };
        
        theirADN.binding = binding;
        us++;
        
        if ((extractFunction) && (binding.templateNode) && (theirADN)) {
            if (binding.templateNode.fieldName) {
                extractFunction(binding.templateNode.fieldName, theirADN.text, theirADN.domNode);
            } else if (binding.templateNode.sampleADN.nodeType == 1) {
                binding.templateNode.bindChildren(theirADN, extractFunction, extractTextFunction);
            } else {
                binding.templateNode.bindText(theirADN, extractTextFunction);
            }
        }
    }
}

TemplateNode.prototype.learnText = function(adns) {
    this.children = [];
    
    var i = 0;
    while (i < adns.length) {
        if (adns[i]) break;
        i++;
    }
    if (i >= adns.length) {
        return;
    }
    
    var textTemplateNodes = map(
        function(token) { 
            return new TextTemplateNode([ token ]); 
        }, 
        PBToken.tokenize(adns[i].text, adns[i].nodeName)
    );
    
    var learn = function(templates, samples, recurse) {
        if (templates.length == 0) {
            return map(
                function(token) { 
                    var node = new TextTemplateNode([ token ]);
                    node.variable = true;
                    return node;
                }, samples
            );
        } else if (samples.length == 0) {
            map(
                function(templateNode) { 
                    templateNode.variable = true;
                    return templateNode;
                }, templates
            );
            return templates;
        }
        
        var bestSampleIndex = -1;
        var bestTemplateIndex = -1;
        var bestTokenCount = -1;
        var bestRunningLength = -1;
        
        for (var i = 0; i < samples.length; i++) {
            var sample = samples[i];
            
            for (var j = 0; j < templates.length; j++) {
                var template = templates[j];
                if (sample.text == template.tokens[0].text) {
                    var runningLength = 0;
                    var tokenCount = 0;
                    
                    for (var k = j; k < templates.length && (i + (k - j)) < samples.length; k++) {
                        var template2 = templates[k];
                        var sample2 = samples[i + (k - j)];
                        //if (sample2.text == template2.tokens[0].text) {
                        if (sample2.type == template2.tokens[0].type &&
                            (sample2.type != 0 || sample2.text == template2.tokens[0].text)) {
                            runningLength += sample2.text.length;
                            tokenCount++;
                        } else {
                            break;
                        }
                    }
                    
                    if (runningLength > bestRunningLength) {
                        bestSampleIndex = i;
                        bestTemplateIndex = j;
                        bestTokenCount = tokenCount;
                        bestRunningLength = runningLength;
                    }
                }
            }
        }
        
        if (bestSampleIndex >= 0) {
            var leftTemplates = templates.slice(0, bestTemplateIndex);
            var midTemplates = templates.slice(bestTemplateIndex, bestTemplateIndex + bestTokenCount);
            var rightTemplates = templates.slice(bestTemplateIndex + bestTokenCount);
            
            var leftSamples = samples.slice(0, bestSampleIndex);
            var midSamples = samples.slice(bestSampleIndex, bestSampleIndex + bestTokenCount);
            var rightSamples = samples.slice(bestSampleIndex + bestTokenCount);
            
            for (var i = 0; i < midTemplates.length; i++) {
                if (midTemplates[i].tokens[0].text != midSamples[i].text) {
                    midTemplates[i].variable = true;
                }
            }
            
            return recurse(leftTemplates, leftSamples, recurse).
                concat(midTemplates).
                concat(recurse(rightTemplates, rightSamples, recurse));
        } else {
            map(
                function(templateNode) { 
                    templateNode.variable = true;
                    return templateNode;
                }, templates
            );
            return templates;
        }
    };
    
    for (var j = i + 1; j < adns.length; j++) {
        if (adns[j]) {
            textTemplateNodes = learn(textTemplateNodes, PBToken.tokenize(trim(adns[j].text), adns[j].nodeName), learn);
        }
    }
    
    /*
     *  Consolidate consecutive tokens
     */
    var i = 0;
    while (i < textTemplateNodes.length) {
        var type = textTemplateNodes[i].tokens[0].type;
        var variable = textTemplateNodes[i].variable;
        
        var tokens = [];
        while (i < textTemplateNodes.length && 
                textTemplateNodes[i].variable == variable && 
                textTemplateNodes[i].tokens[0].type == type) {
                
            tokens.push(textTemplateNodes[i].tokens[0]);
            i++;
            
            if (type != 0) {
                break;
            }
        }
        var newTextTemplateNode = new TextTemplateNode(tokens);
        newTextTemplateNode.variable = variable;
        
        this.children.push(newTextTemplateNode);
    }
}

TemplateNode.prototype.bindText = function(adn, extractTextFunction) {
    var bindings = [];
    
    var textTemplateNodes = this.children;
    if (textTemplateNodes.length == 0) {
        return bindings;
    }
    
    var textFragments = [];
    var fieldNames = [];
    var fieldValues = [];
    
    var tokens = PBToken.tokenize(adn.text, adn.nodeName, true /* keep whitespace */);
    if (!(tokens instanceof Array)) {
        return bindings;
    }
    
    for (var i = 0; i < textTemplateNodes.length; i++) {
        var textTemplateNode = textTemplateNodes[i];
        var binding = null;
        
        /*
         *  Trim whitespace first
         */
        var whitespace = "";
        while (tokens.length > 0 && tokens[0].type == PB_TOKEN_WHITESPACE) {
            whitespace += tokens[0].text;
            tokens = tokens.slice(1);
        }
        if (whitespace.length > 0) {
            textFragments.push(whitespace);
            fieldNames.push("");
            fieldValues.push(null);
        }
        
        if (tokens.length > 0) {
            if (textTemplateNode.variable) {
                if (textTemplateNode.tokens[0].type == PB_TOKEN_TEXT) { // string
                    /*
                     *  A variable template node of strings don't have any 
                     *  fixed count of tokens. We need to find the next
                     *  template token, if any.
                     */
                    if (i + 1 < textTemplateNodes.length) {
                        var nextTextTemplateNode = textTemplateNodes[i + 1];
                        var k = 1;
                        if (nextTextTemplateNode.variable) {
                            while (k < tokens.length) {
                                if (tokens[k].type != PB_TOKEN_TEXT && 
                                    tokens[k].type != PB_TOKEN_WHITESPACE) {
                                    break;
                                }
                                k++;
                            }
                        } else {
                            while (k < tokens.length) {
                                if ((tokens[k].type != PB_TOKEN_TEXT && tokens[k].type != PB_TOKEN_WHITESPACE) || 
                                    tokens[k].text == nextTextTemplateNode.tokens[0].text) {
                                    break;
                                }
                                k++;
                            }
                        }
                        binding = {
                            text: map(function(t) { return t.text; }, tokens.slice(0, k)).join("")
                        };
                        tokens = tokens.slice(k);
                    } else {
                        /*
                         *  All subsequent tokens are variable.
                         */
                        binding = {
                            text: map(function(t) { return t.text; }, tokens).join("")
                        };
                        tokens = [];
                    }
                } else {
                    if (tokens[0].type == textTemplateNode.tokens[0].type) {
                        binding = {
                            text: tokens[0].text,
                            value: tokens[0].value
                        };
                        tokens = tokens.slice(1);
                    }
                }
            } else {
                var k = 0;
                var tokens2 = [];
                while (tokens.length > 0 && k < textTemplateNode.tokens.length) {
                    if (tokens[0].type == PB_TOKEN_WHITESPACE) {
                        tokens2.push(tokens[0]);
                        tokens = tokens.slice(1);
                    } else if (
                        tokens[0].type == textTemplateNode.tokens[k].type &&
                        tokens[0].text == textTemplateNode.tokens[k].text) {
                        
                        tokens2.push(tokens[0]);
                        tokens = tokens.slice(1);
                        k++;
                    } else {
                        break;
                    }
                }
                
                binding = {
                    text: map(function(t) { return t.text; }, tokens2).join("")
                }
            }
            
            if (binding) {
                textFragments.push(binding.text);
                fieldValues.push(binding.value ? binding.value : null);
                fieldNames.push((textTemplateNode.fieldName) ? textTemplateNode.fieldName : "");
            }
        }
        bindings.push(binding);
    }
    
    for (var i = 0; i < tokens.length; i++) {
        textFragments.push(tokens[i].text);
        fieldNames.push("");
        fieldValues.push(null);
    }
    
    if (extractTextFunction) {
        extractTextFunction(textFragments, fieldNames, fieldValues, adn.domNode);
    }
    
    return bindings;
}

/*----------------------------------------------------------------------
 *  TextTemplateNode
 *----------------------------------------------------------------------
 */
 
function TextTemplateNode(tokens) {
    this.tokens = tokens;
    this.variable = false;
    this.hits = 0;
    this.misses = 0;
}
 
/*----------------------------------------------------------------------
 *  Template
 *----------------------------------------------------------------------
 */
 
function Template(templateNodes) {
    this.children = templateNodes;
}

Template.prototype.bind = function(item, extractFunction, extractTextFunction) {
    for (var i = 0; i < item.children.length; i++) {
        this.children[i].bindChildren(item.children[i], extractFunction, extractTextFunction);
    }
}

Template.createTemplateFromItems = function(items) {
    var template = items[0].createTemplate();
    for (var i = 0; i < template.children.length; i++) {
        var adns = map(
            function(item) {
                return (i < item.children.length) ? item.children[i] : null;
            },
            items
        );
        template.children[i].learnDeep(adns);
    }
    
    return template;
}

Template.prototype.improveTemplateFromItems = function(items) {
    for (var i = 0; i < this.children.length; i++) {
        var adns = map(
            function(item) {
                return (i < item.children.length) ? item.children[i] : null;
            },
            items
        );
        this.children[i].learnDeep(adns);
    }
}

Template.prototype.assignExtractableFields = function() {
    var o = { fieldIndex : 0 };
    for (var i = 0; i < this.children.length; i++) {
        this.children[i].assignExtractableFields(o, 0);
    }
}

Template._learn = function(
    templateItems, 
    sampleItems, 
    costFunction, 
    compareFunction, 
    hitFunction,
    missFunction,
    constructorFunction
) {
    var us = 0;
    var them = 0;
    
    while (us < templateItems.length && them < sampleItems.length) {
        var templateItem = templateItems[us];
        var sampleItem = sampleItems[them];
        
        /*
         *  Find their best that matches ours
         */
        var theirBestOffset = 0;
        var theirMinCost = 1000;
        for (var i = them; i < sampleItems.length; i++) {
            var c = costFunction(templateItem, sampleItems[i]) * (theirBestOffset + 1);
            if (c < theirMinCost) {
                theirBestOffset = i - them;
                theirMinCost = c;
            }
        }
        
        /*
         *  Find our best that matches theirs
         */
        var ourBestOffset = 0;
        var ourMinCost = 1000;
        for (var i = us; i < templateItems.length; i++) {
            var c = costFunction(templateItems[i], sampleItem) * (ourBestOffset + 1);
            if (c < ourMinCost) {
                ourBestOffset = i - us;
                ourMinCost = c;
            }
        }
        
        if (theirMinCost < ourMinCost) {
            if (compareFunction(templateItem, sampleItems[them + theirBestOffset]) != 0) {
                missFunction(templateItem);
            } else {
                hitFunction(templateItem);
            }
            
            /*
             *  Insert new template nodes from them and mark optional
             */
            var args = [ us, 0 ];
            for (var offset = 0; offset < theirBestOffset; offset++) {
                var templateItem = constructorFunction(sampleItems[them + offset]);
                missFunction(templateItem);
                
                args.push(templateItem);
            }
            templateItems.splice.apply(templateItems, args);
            
            us += theirBestOffset + 1; // the insertion affects our offset, too
            them += theirBestOffset + 1;
        } else {
            if (compareFunction(templateItems[us + ourBestOffset], sampleItem) != 0) {
                missFunction(templateItems[us + ourBestOffset]);
            } else {
                hitFunction(templateItems[us + ourBestOffset]);
            }
            
            /*
             *  Mark some of our nodes as optional
             */
            for (var offset = 0; offset < ourBestOffset; offset++) {
                missFunction(templateItems[us + offset]);
            }
            
            us += ourBestOffset + 1;
            them += 1;
        }
    }
    
    if (us < templateItems.length) {
        /*
         *  The rest of our nodes are optional
         */
        while (us < templateItems.length) {
            missFunction(templateItems[us]);
            us++;
        }
    } else if (them < sampleItems.length) {
        /*
         *  Insert new template nodes from them and mark optional
         */
        while (them < sampleItems.length) {
            var templateItem = constructorFunction(sampleItems[them]);
            missFunction(templateItem);
            
            templateItems.push(templateItem);
            
            them++;
        }
    }
}
