/*----------------------------------------------------------------------
 * The nsLogger Module
 *----------------------------------------------------------------------
 */

var LoggerModule = {
    myCID : Components.ID("{51865cf0-1445-11da-8cd6-0800200c9a66}"),
    myProgID : "@simile.mit.edu/logger;1",

    registerSelf : function(compMgr, fileSpec, location, type) {
        compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
        compMgr.registerFactoryLocation(
            this.myCID,
            "Logging Services",
            this.myProgID,
            fileSpec,
            location,
            type
        );
    },

    unregisterSelf : function(compMgr, fileSpec, location) {
        compMgr = compMgr.QueryInterface(nsIComponentRegistrar);
        compMgr.unregisterFactoryLocation(this.myCID, fileSpec);
    },

    getClassObject : function(compMgr, cid, iid) {
        if (!cid.equals(this.myCID)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        if (!iid.equals(Components.interfaces.nsIFactory)) {
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        }
        return this.myFactory;
    },

    canUnload : function(compMgr) {
        Logger_singleton = null;        
        return true;
    },

    myFactory : {
        createInstance : function(outer, iid) {
            if (outer != null) {
                throw Components.results.NS_ERROR_NO_AGGREGATION;
            }

            if (!Logger_singleton) {
                Logger_singleton = new Logger();
            }

            return Logger_singleton.QueryInterface(iid);
        }
    }
}

function NSGetModule(compMgr, fileSpec) {
    return LoggerModule;
}

/*----------------------------------------------------------------------
 * The Logger Component
 *----------------------------------------------------------------------
 */

var Logger_singleton = null;

function Logger() {};

Logger.prototype = {

    _system : true,
    _enabled : false,
    _redirecting : false,
    _previousTime : 0,
    _indentation : 0,

    QueryInterface : function(iid) {
        if (!iid.equals(Components.interfaces.nsILogger) &&
            !iid.equals(Components.interfaces.nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    },
    
    enable : function (enabled) {
        this._setDebugMode(enabled);
        this._enabled = enabled;
    },
    isEnabled : function() {
        return this._enabled;
    },

    enableTracing : function(enabled) {
        this._tracing = enabled;
    },
    tracing : function() {
        return this._tracing;
    },

    logToSystemConsole : function() {
        this._system = true;
    },
    logToJavascriptConsole : function() {
        this._system = false;
    },
    
    trace : function(context,msg) {
        if (this._enabled && this.tracing()) {
            this._log(context,msg);
        }
    },
    log : function(context,msg) {
        if (this._enabled) {
            this._log(context,msg);
        }
    },

    redirectJSConsole : function (redirect) {
        if (redirect && this._enabled && !this._redirecting) {
            var debug = this;
            
            this._consoleObserver = {
                observe : function (msg) {
                    debug.log("JSConsole",msg.message);
                },
                
                QueryInterface : function (iid) {
                    if (!iid.equals(Components.interfaces.nsIConsoleListener) &&
                        !iid.equals(Components.interfaces.nsISupports)) {
                        throw Components.results.NS_ERROR_NO_INTERFACE;
                    }
                    return this;
                }
            };
            
            Components.classes["@mozilla.org/consoleservice;1"]
                .getService(Components.interfaces.nsIConsoleService)
                    .registerListener(this._consoleObserver);
                    
            this._redirecting = true;
        } else if (this._consoleObserver) {
            Components.classes["@mozilla.org/consoleservice;1"]
                .getService(Components.interfaces.nsIConsoleService)
                    .unregisterListener(this._consoleObserver);
            this._redirecting = false;
        }
    },

    // NOTE(SM): This code is not built for speed and memory consumption
    // in mind, therefore is not very fast and generates tons of garbage, 
    // so make sure you disable logging when you don't need it.
    
    _log : function (context,message) {
        var leader = message.charAt(0);
        if ((leader == '<') && (this._indentation > 0)) this._indentation--;

        var msg = '';
        var d = new Date();
        var delta = 0;
        if (this._previousTime > 0) {
            delta = d.getTime() - this._previousTime;
        }
        this._previousTime = d.getTime();

        if ((this._previousTime == 0) || (delta > 3000)) {
            msg += "\n";
            this._indentation = 0; // reset indentation after a while, as we might have runaway/unmatched log entries
        }
        msg += this._pad(d.getHours(),2) + ":" + this._pad(d.getMinutes(),2) + ":" + this._pad(d.getSeconds(),2) + "." + this._pad(d.getMilliseconds(),3);
        this._previousTime = d.getTime();
        msg += " [" + context.substr(0,15) + this._dynastr(" ", 15 - context.length) + "] " + this._dynastr("  ", this._indentation) + message;
        if (delta < 3000) {
            msg += " (" + delta + "ms)";
        }

        if (this._system) {
            dump(msg);
            dump("\n");
        } else {
            Components.classes["@mozilla.org/consoleservice;1"]
                .getService(Components.interfaces.nsIConsoleService)
                    .logStringMessage(msg);
        }

        if (leader == '>') this._indentation++;
    },

    _pad : function (number, digits) {
        var str = number;
        for (var i = digits - 1 ; i > 0 ; i--) {
            if (number < Math.pow(10,i)) {
                str = "0" + str;
            }
        }
        return str;
    },

    _dynastr : function (str,count) {
        var result = "";
        for (var i = 0 ; i < count ; i++) {
            result = str + result;
        }
        return result;
    },
    
    _setDebugMode : function (enabled) {
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
        prefs.setBoolPref("browser.dom.window.dump.enabled", enabled);
        prefs.setBoolPref("javascript.options.showInConsole", enabled);
        prefs.setBoolPref("nglayout.debug.disable_xul_cache", enabled);
        prefs.setBoolPref("javascript.options.strict", enabled);
    },
    
    // NOTE(SM): Blatantly stolen, ehm borrowed :-), from Ted Mielczarek's outstanding 
    // "extensiondev" firefox extension (check it out at 
    // http://ted.mielczarek.org/code/mozilla/extensiondev/)
    // Thanks! You rock!
    
    introspect : function (object) {
    
        var names = ["Methods", "Fields", "Unreachables"];
        var results = [[], [], []]; // array of (empty) arrays of arrays!
        var p, j, i; // loop variables, several used multiple times
        
        var protoLevels = 0;
        
        for (p = object; p; p = p.__proto__) {
            for (i = 0; i < names.length; ++i) {
                results[i][protoLevels] = [];
            }
            ++protoLevels;
        }
        
        for (var value in object) {
            // Shortcoming: doesn't check that VALUES are the same in object and prototype.
            
            var protoLevel = -1;
            try {
                for (p = object; p && (value in p); p = p.__proto__) {
                    ++protoLevel;
                }
            } catch (e) { 
                protoLevel = 0; 
            } // "in" operator throws when param to props() is a string
            
            var type = 1;
            try {
                if ((typeof object[value]) == "function") {
                    type = 0;
                }
            } catch (e) { 
                type = 2; 
            }
            
            results[type][protoLevel].push(value);
        }
    
        function times(s, n) { 
            return n ? s + times(s, n-1) : ""; 
        }
    
        var str = "";
        
        for (j = 0; j < protoLevels; ++j) {
            for (i = 0; i < names.length; ++i) {
                if (results[i][j].length) {
                    str += names[i] + times(" of prototype", j) + "\n  " + results[i][j].join("\n  ") + "\n";
                }
            }
        }
        
        return str;
    }

};
